from __future__ import annotations

import copy
import shutil
from pathlib import Path
from typing import TYPE_CHECKING

import platformdirs
import yaml

from cyberdrop_dl.managers.path_manager import APP_STORAGE, DOWNLOAD_STORAGE
from cyberdrop_dl.utils.transfer.transfer_v4_db import transfer_v4_db

if TYPE_CHECKING:
    from cyberdrop_dl.managers.manager import Manager


class TransitionManager:
    def __init__(self, manager: Manager):
        self.manager = manager

    def startup(self) -> None:
        """Startup"""
        OLD_APP_STORAGE: Path = Path(platformdirs.user_config_dir("Cyberdrop-DL"))
        OLD_DOWNLOAD_STORAGE = Path(platformdirs.user_downloads_path()) / "Cyberdrop-DL Downloads"

        if APP_STORAGE.exists():
            cache_file = APP_STORAGE / "Cache" / "cache.yaml"
            if cache_file.is_file() and self.check_cache_for_moved(cache_file):
                return

        OLD_FILES = Path("./Old Files")
        OLD_FILES.mkdir(parents=True, exist_ok=True)

        if OLD_APP_STORAGE.exists():
            if APP_STORAGE.exists():
                if APP_STORAGE.with_name("AppData_OLD").exists():
                    APP_STORAGE.rename(APP_STORAGE.with_name("AppData_OLD2"))
                APP_STORAGE.rename(APP_STORAGE.with_name("AppData_OLD"))
            shutil.copytree(OLD_APP_STORAGE, APP_STORAGE, dirs_exist_ok=True)
            shutil.rmtree(OLD_APP_STORAGE)

        if OLD_DOWNLOAD_STORAGE.exists():
            shutil.copytree(OLD_DOWNLOAD_STORAGE, DOWNLOAD_STORAGE, dirs_exist_ok=True)
            shutil.rmtree(OLD_DOWNLOAD_STORAGE)

        if Path("./download_history.sqlite").is_file():
            transfer_v4_db(Path("./download_history.sqlite"), APP_STORAGE / "Cache" / "cyberdrop.db")
            Path("./download_history.sqlite").rename(OLD_FILES / "download_history1.sqlite")

        if (APP_STORAGE / "download_history.sqlite").is_file():
            transfer_v4_db(APP_STORAGE / "download_history.sqlite", APP_STORAGE / "Cache" / "cyberdrop.db")
            (APP_STORAGE / "download_history.sqlite").rename(OLD_FILES / "download_history2.sqlite")

        if Path("./config.yaml").is_file():
            try:
                self.transfer_v4_config(Path("./config.yaml"), "Imported V4")
                self.update_default_config(APP_STORAGE / "Cache" / "cache.yaml", "Imported V4")
            except Exception as e:
                pass
            Path("./config.yaml").rename(OLD_FILES / "config.yaml")

        if Path("./downloader.log").is_file():
            Path("./downloader.log").rename(OLD_FILES / "downloader.log")
        if Path("./Errored_Download_URLs.csv").is_file():
            Path("./Errored_Download_URLs.csv").rename(OLD_FILES / "Errored_Download_URLs.csv")
        if Path("./Errored_Scrape_URLs.csv").is_file():
            Path("./Errored_Scrape_URLs.csv").rename(OLD_FILES / "Errored_Scrape_URLs.csv")
        if Path("./Unsupported_URLs.csv").is_file():
            Path("./Unsupported_URLs.csv").rename(OLD_FILES / "Unsupported_URLs.csv")

        self.update_cache(APP_STORAGE / "Cache" / "cache.yaml")
        pass

    def check_cache_for_moved(self, cache_file: Path) -> bool:
        """Checks the cache for moved files"""
        with open(cache_file, 'r') as yaml_file:
            cache = yaml.load(yaml_file.read(), Loader=yaml.FullLoader)
        if cache is None:
            with open(cache_file, 'w') as yaml_file:
                cache = {"first_startup_completed": False}
                yaml.dump(cache, yaml_file)
        moved = bool(cache.get("first_startup_completed", False))
        return moved

    def update_cache(self, cache_file: Path) -> None:
        """Updates the cache to reflect the new location"""
        cache_file.parent.mkdir(parents=True, exist_ok=True)
        cache_file.touch(exist_ok=True)
        with open(cache_file, 'r') as yaml_file:
            cache = yaml.load(yaml_file.read(), Loader=yaml.FullLoader)
        if cache is None:
            cache = {"first_startup_completed": False}
        cache['first_startup_completed'] = True
        with open(cache_file, 'w') as yaml_file:
            yaml.dump(cache, yaml_file)

    def update_default_config(self, cache_file: Path, config_name: str) -> None:
        """Updates the default config in the cache"""
        cache_file.parent.mkdir(parents=True, exist_ok=True)
        cache_file.touch(exist_ok=True)
        with open(cache_file, 'r') as yaml_file:
            cache = yaml.load(yaml_file.read(), Loader=yaml.FullLoader)
        if cache is None:
            cache = {"first_startup_completed": False}
        cache['default_config'] = config_name
        with open(cache_file, 'w') as yaml_file:
            yaml.dump(cache, yaml_file)

    def transfer_v4_config(self, old_config_path: Path, new_config_name: str) -> None:
        """Transfers a V4 config into V5 possession"""
        from cyberdrop_dl.utils.args.config_definitions import settings, authentication_settings, global_settings
        new_auth_data = copy.deepcopy(authentication_settings)
        new_user_data = copy.deepcopy(settings)
        new_global_settings_data = copy.deepcopy(global_settings)

        if Path("./URLs.txt").is_file():
            new_user_data['Files']['input_file'] = Path("./URLs.txt")
        else:
            new_user_data['Files']['input_file'] = APP_STORAGE / "Configs" / new_config_name / "URLs.txt"
        new_user_data['Files']['download_folder'] = DOWNLOAD_STORAGE
        new_user_data["Logs"]["log_folder"] = APP_STORAGE / "Configs" / new_config_name / "Logs"
        new_user_data['Sorting']['sort_folder'] = DOWNLOAD_STORAGE / "Cyberdrop-DL Sorted Downloads"

        with open(old_config_path, 'r') as yaml_file:
            old_data = yaml.load(yaml_file.read(), Loader=yaml.FullLoader)
        old_data = old_data['Configuration']

        # Auth data transfer
        new_auth_data['Forums']['nudostar_username'] = old_data['Authentication']['nudostar_username']
        new_auth_data['Forums']['nudostar_password'] = old_data['Authentication']['nudostar_password']
        new_auth_data['Forums']['simpcity_username'] = old_data['Authentication']['simpcity_username']
        new_auth_data['Forums']['simpcity_password'] = old_data['Authentication']['simpcity_password']
        new_auth_data['Forums']['socialmediagirls_username'] = old_data['Authentication']['socialmediagirls_username']
        new_auth_data['Forums']['socialmediagirls_password'] = old_data['Authentication']['socialmediagirls_password']
        new_auth_data['Forums']['xbunker_username'] = old_data['Authentication']['xbunker_username']
        new_auth_data['Forums']['xbunker_password'] = old_data['Authentication']['xbunker_password']

        new_auth_data['JDownloader']['jdownloader_username'] = old_data['JDownloader']['jdownloader_username']
        new_auth_data['JDownloader']['jdownloader_password'] = old_data['JDownloader']['jdownloader_password']
        new_auth_data['JDownloader']['jdownloader_device'] = old_data['JDownloader']['jdownloader_device']

        new_auth_data['Reddit']['reddit_personal_use_script'] = old_data['Authentication']['reddit_personal_use_script']
        new_auth_data['Reddit']['reddit_secret'] = old_data['Authentication']['reddit_secret']

        new_auth_data['GoFile']['gofile_api_key'] = old_data['Authentication']['gofile_api_key']
        new_auth_data['Imgur']['imgur_client_id'] = old_data['Authentication']['imgur_client_id']
        new_auth_data['PixelDrain']['pixeldrain_api_key'] = old_data['Authentication']['pixeldrain_api_key']

        # User data transfer
        new_user_data['Download_Options']['block_download_sub_folders'] = old_data['Runtime']['block_sub_folders']
        new_user_data['Download_Options']['disable_download_attempt_limit'] = old_data['Runtime']['disable_attempt_limit']
        new_user_data['Download_Options']['include_album_id_in_folder_name'] = old_data['Runtime']['include_id']
        new_user_data['Download_Options']['remove_generated_id_from_filenames'] = old_data['Runtime']['remove_bunkr_identifier']
        new_user_data['Download_Options']['separate_posts'] = old_data['Forum_Options']['separate_posts']
        new_user_data['Download_Options']['skip_download_mark_completed'] = False

        new_user_data['File_Size_Limits']['maximum_image_size'] = old_data['Runtime']['filesize_maximum_images']
        new_user_data['File_Size_Limits']['maximum_other_size'] = old_data['Runtime']['filesize_maximum_other']
        new_user_data['File_Size_Limits']['maximum_video_size'] = old_data['Runtime']['filesize_maximum_videos']
        new_user_data['File_Size_Limits']['minimum_image_size'] = old_data['Runtime']['filesize_minimum_images']
        new_user_data['File_Size_Limits']['minimum_other_size'] = old_data['Runtime']['filesize_minimum_other']
        new_user_data['File_Size_Limits']['minimum_video_size'] = old_data['Runtime']['filesize_minimum_videos']

        new_user_data['Ignore_Options']['exclude_videos'] = old_data['Ignore']['exclude_videos']
        new_user_data['Ignore_Options']['exclude_images'] = old_data['Ignore']['exclude_images']
        new_user_data['Ignore_Options']['exclude_other'] = old_data['Ignore']['exclude_other']
        new_user_data['Ignore_Options']['exclude_audio'] = old_data['Ignore']['exclude_audio']
        new_user_data['Ignore_Options']['ignore_coomer_ads'] = old_data['Ignore']['skip_coomer_ads']
        new_user_data['Ignore_Options']['skip_hosts'] = old_data['Ignore']['skip_hosts']
        new_user_data['Ignore_Options']['only_hosts'] = old_data['Ignore']['only_hosts']

        new_user_data['Runtime_Options']['ignore_history'] = old_data['Ignore']['ignore_history']
        new_user_data['Runtime_Options']['skip_check_for_partial_files'] = old_data['Runtime']['skip_check_for_partial_files_and_empty_dirs']
        new_user_data['Runtime_Options']['skip_check_for_empty_folders'] = old_data['Runtime']['skip_check_for_partial_files_and_empty_dirs']
        new_user_data['Runtime_Options']['send_unsupported_to_jdownloader'] = old_data['JDownloader']['apply_jdownloader']

        new_user_data['Sorting']['sort_downloads'] = old_data['Sorting']['sort_downloads']

        if Path("./URLs.txt").is_file():
            new_user_data['Files']['input_file'] = "./URLs.txt"

        # Global data transfer
        new_global_settings_data['General']['allow_insecure_connections'] = old_data['Runtime']['allow_insecure_connections']
        new_global_settings_data['General']['user_agent'] = old_data['Runtime']['user_agent']
        new_global_settings_data['General']['proxy'] = old_data['Runtime']['proxy']
        new_global_settings_data['General']['max_file_name_length'] = old_data['Runtime']['max_filename_length']
        new_global_settings_data['General']['max_folder_name_length'] = old_data['Runtime']['max_folder_name_length']
        new_global_settings_data['General']['required_free_space'] = old_data['Runtime']['required_free_space']

        new_global_settings_data['Rate_Limiting_Options']['connection_timeout'] = old_data['Ratelimiting']['connection_timeout']
        new_global_settings_data['Rate_Limiting_Options']['download_attempts'] = old_data['Runtime']['attempts']
        new_global_settings_data['Rate_Limiting_Options']['download_delay'] = old_data['Ratelimiting']['throttle']
        new_global_settings_data['Rate_Limiting_Options']['read_timeout'] = old_data['Ratelimiting']['read_timeout']
        new_global_settings_data['Rate_Limiting_Options']['rate_limit'] = old_data['Ratelimiting']['ratelimit']
        new_global_settings_data['Rate_Limiting_Options']['max_simultaneous_downloads_per_domain'] = old_data['Runtime']['max_concurrent_downloads_per_domain']

        new_user_data['Files']['input_file'] = str(new_user_data['Files']['input_file'])
        new_user_data['Files']['download_folder'] = str(new_user_data['Files']['download_folder'])
        new_user_data["Logs"]["log_folder"] = str(new_user_data["Logs"]["log_folder"])
        new_user_data['Sorting']['sort_folder'] = str(new_user_data['Sorting']['sort_folder'])

        # Write config
        new_config_path = APP_STORAGE / "Configs" / new_config_name / "settings.yaml"
        new_config_path.parent.mkdir(parents=True, exist_ok=True)
        with open(new_config_path, 'w') as yaml_file:
            yaml.dump(new_user_data, yaml_file)
        new_auth_path = APP_STORAGE / "Configs" / "authentication.yaml"
        with open(new_auth_path, 'w') as yaml_file:
            yaml.dump(new_auth_data, yaml_file)
        new_global_settings_path = APP_STORAGE / "Configs" / "global_settings.yaml"
        with open(new_global_settings_path, 'w') as yaml_file:
            yaml.dump(new_global_settings_data, yaml_file)



from __future__ import annotations

from typing import TYPE_CHECKING

from aiolimiter import AsyncLimiter
from yarl import URL

from cyberdrop_dl.clients.errors import ScrapeFailure, FailedLoginFailure
from cyberdrop_dl.scraper.crawler import Crawler
from cyberdrop_dl.utils.dataclasses.url_objects import ScrapeItem
from cyberdrop_dl.utils.utilities import error_handling_wrapper, log, get_filename_and_ext

if TYPE_CHECKING:
    from cyberdrop_dl.managers.manager import Manager


class ImgurCrawler(Crawler):
    def __init__(self, manager: Manager):
        super().__init__(manager, "imgur", "Imgur")
        self.imgur_api = URL("https://api.imgur.com/3/")
        self.imgur_client_id = self.manager.config_manager.authentication_data["Imgur"]["imgur_client_id"]
        self.imgur_client_remaining = 12500
        self.headers = {"Authorization": f"Client-ID {self.imgur_client_id}"}
        self.request_limiter = AsyncLimiter(10, 1)

    """~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""

    async def fetch(self, scrape_item: ScrapeItem) -> None:
        """Determines where to send the scrape item based on the url"""
        task_id = await self.scraping_progress.add_task(scrape_item.url)

        if "i.imgur.com" in scrape_item.url.host:
            await self.handle_direct(scrape_item)
        elif "a" in scrape_item.url.parts:
            await self.album(scrape_item)
        else:
            await self.image(scrape_item)

        await self.scraping_progress.remove_task(task_id)

    @error_handling_wrapper
    async def album(self, scrape_item: ScrapeItem) -> None:
        """Scrapes an album"""
        if self.imgur_client_id == "":
            await log("To scrape imgur content, you need to provide a client id", 30)
            raise FailedLoginFailure(status=401, message="No Imgur Client ID provided")
        await self.check_imgur_credits()

        album_id = scrape_item.url.parts[-1]

        async with self.request_limiter:
            JSON_Obj = await self.client.get_json(self.domain, self.imgur_api / f"album/{album_id}", headers_inc=self.headers)
        title_part = JSON_Obj["data"].get("title", album_id)
        title = await self.create_title(title_part, scrape_item.url.parts[2], None)

        async with self.request_limiter:
            JSON_Obj = await self.client.get_json(self.domain, self.imgur_api / f"album/{album_id}/images", headers_inc=self.headers)

        for image in JSON_Obj["data"]:
            link = URL(image["link"])
            date = image["datetime"]
            new_scrape_item = await self.create_scrape_item(scrape_item, link, title, True, date)
            await self.handle_direct(new_scrape_item)

    @error_handling_wrapper
    async def image(self, scrape_item: ScrapeItem) -> None:
        """Scrapes an image"""
        if self.imgur_client_id == "":
            await log("To scrape imgur content, you need to provide a client id", 30)
            raise FailedLoginFailure(status=401, message="No Imgur Client ID provided")
        await self.check_imgur_credits()

        image_id = scrape_item.url.parts[-1]
        async with self.request_limiter:
            JSON_Obj = await self.client.get_json(self.domain, self.imgur_api / f"image/{image_id}", headers_inc=self.headers)

        date = JSON_Obj["data"]["datetime"]
        link = URL(JSON_Obj["data"]["link"])
        new_scrape_item = await self.create_scrape_item(scrape_item, link, "", True, date)
        await self.handle_direct(new_scrape_item)

    @error_handling_wrapper
    async def handle_direct(self, scrape_item: ScrapeItem) -> None:
        """Scrapes an image"""
        filename, ext = await get_filename_and_ext(scrape_item.url.name)
        if ext.lower() == ".gifv" or ext.lower() == ".mp4":
            filename = filename.replace(ext, ".mp4")
            ext = ".mp4"
            scrape_item.url = URL("https://imgur.com/download") / filename.replace(ext, "")
        await self.handle_file(scrape_item.url, scrape_item, filename, ext)

    """~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""

    async def check_imgur_credits(self) -> None:
        """Checks the remaining credits"""
        credits_obj = await self.client.get_json(self.domain, self.imgur_api / "credits", headers_inc=self.headers)
        self.imgur_client_remaining = credits_obj["data"]["ClientRemaining"]
        if self.imgur_client_remaining < 100:
            raise ScrapeFailure(429, "Imgur API rate limit reached")



import sqlite3
from pathlib import Path

from cyberdrop_dl.utils.database.table_definitions import create_history, create_temp


def transfer_v4_db(db_path: Path, new_db_path: Path) -> None:
    """Transfers a V4 database into V5 possession"""
    old_db_connection = sqlite3.connect(db_path)
    if not new_db_path.exists():
        new_db_path.parent.mkdir(parents=True, exist_ok=True)
        new_db_path.touch()
    new_db_connection = sqlite3.connect(new_db_path)

    new_db_connection.execute(create_history)
    new_db_connection.execute(create_temp)

    query = "SELECT domain, url_path, referer, download_path, download_filename, original_filename, completed FROM media WHERE completed = 1"
    old_data_history = old_db_connection.execute(query).fetchall()

    old_data_revised = []
    for row in old_data_history:
        row = list(row)
        row[3] = str(Path(row[3]).parent)
        old_data_revised.append(tuple(row))

    new_db_connection.executemany("insert or ignore into media values (?, ?, ?, ?, ?, ?, ?)", old_data_revised)
    del old_data_history

    new_db_connection.commit()
    old_db_connection.close()
    new_db_connection.close()



from __future__ import annotations

import asyncio
import contextlib
import shutil
from base64 import b64encode
from typing import TYPE_CHECKING

from cyberdrop_dl.utils.utilities import FILE_FORMATS

if TYPE_CHECKING:
    from typing import Dict

    from cyberdrop_dl.managers.manager import Manager
    from cyberdrop_dl.utils.dataclasses.url_objects import MediaItem


class FileLock:
    """Is this necessary? No. But I want it."""
    def __init__(self):
        self._locked_files = {}

    async def check_lock(self, filename: str) -> None:
        """Checks if the file is locked"""
        try:
            await self._locked_files[filename].acquire()
        except KeyError:
            self._locked_files[filename] = asyncio.Lock()
            await self._locked_files[filename].acquire()

    async def release_lock(self, filename: str) -> None:
        """Releases the file lock"""
        with contextlib.suppress(KeyError, RuntimeError):
            self._locked_files[filename].release()


class DownloadManager:
    def __init__(self, manager: Manager):
        self.manager = manager
        self._download_instances: Dict = {}

        self.file_lock = FileLock()

        self.download_limits = {'bunkr': 1, 'bunkrr': 1, 'cyberdrop': 1, 'coomer': 2, 'cyberfile': 2, 'kemono': 2, "pixeldrain": 2}

    async def get_download_limit(self, key: str) -> int:
        """Returns the download limit for a domain"""
        if key in self.download_limits:
            instances = self.download_limits[key]
        else:
            instances = self.manager.config_manager.global_settings_data['Rate_Limiting_Options']['max_simultaneous_downloads_per_domain']

        if instances > self.manager.config_manager.global_settings_data['Rate_Limiting_Options']['max_simultaneous_downloads_per_domain']:
            instances = self.manager.config_manager.global_settings_data['Rate_Limiting_Options']['max_simultaneous_downloads_per_domain']
        return instances

    async def basic_auth(self, username, password) -> str:
        """Returns a basic auth token"""
        token = b64encode(f"{username}:{password}".encode('utf-8')).decode("ascii")
        return f'Basic {token}'

    async def check_free_space(self) -> bool:
        """Checks if there is enough free space on the drive to continue operating"""
        free_space = shutil.disk_usage(self.manager.path_manager.download_dir.parent).free
        free_space_gb = free_space / 1024 ** 3
        return free_space_gb >= self.manager.config_manager.global_settings_data['General']['required_free_space']

    async def check_allowed_filetype(self, media_item: MediaItem) -> bool:
        """Checks if the file type is allowed to download"""
        if media_item.ext in FILE_FORMATS['Images'] and self.manager.config_manager.settings_data['Ignore_Options']['exclude_images']:
            return False
        if media_item.ext in FILE_FORMATS['Videos'] and self.manager.config_manager.settings_data['Ignore_Options']['exclude_videos']:
            return False
        if media_item.ext in FILE_FORMATS['Audio'] and self.manager.config_manager.settings_data['Ignore_Options']['exclude_audio']:
            return False
        if (self.manager.config_manager.settings_data['Ignore_Options']['exclude_other'] and
              media_item.ext not in FILE_FORMATS['Images'] and media_item.ext not in FILE_FORMATS['Videos'] and
              media_item.ext not in FILE_FORMATS['Audio']):
            return False
        return True



from __future__ import annotations

from typing import TYPE_CHECKING

import aiohttp
import asyncpraw
import asyncprawcore
from aiolimiter import AsyncLimiter
from yarl import URL

from cyberdrop_dl.clients.errors import ScrapeFailure, NoExtensionFailure
from cyberdrop_dl.scraper.crawler import Crawler
from cyberdrop_dl.utils.dataclasses.url_objects import ScrapeItem
from cyberdrop_dl.utils.utilities import error_handling_wrapper, log, get_filename_and_ext

if TYPE_CHECKING:
    from cyberdrop_dl.managers.manager import Manager


class RedditCrawler(Crawler):
    def __init__(self, manager: Manager):
        super().__init__(manager, "reddit", "Reddit")
        self.reddit_personal_use_script = self.manager.config_manager.authentication_data['Reddit']['reddit_personal_use_script']
        self.reddit_secret = self.manager.config_manager.authentication_data['Reddit']['reddit_secret']
        self.request_limiter = AsyncLimiter(5, 1)

    """~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""

    async def fetch(self, scrape_item: ScrapeItem) -> None:
        """Determines where to send the scrape item based on the url"""
        task_id = await self.scraping_progress.add_task(scrape_item.url)

        if not self.reddit_personal_use_script or not self.reddit_secret:
            await log("Reddit API credentials not found. Skipping.", 30)
            await self.manager.progress_manager.scrape_stats_progress.add_failure("Failed Login")
            await self.scraping_progress.remove_task(task_id)
            return

        async with aiohttp.ClientSession() as reddit_session:
            reddit = asyncpraw.Reddit(client_id=self.reddit_personal_use_script,
                                      client_secret=self.reddit_secret,
                                      user_agent="CyberDrop-DL",
                                      requestor_kwargs={"session": reddit_session},
                                      check_for_updates=False)

            if "user" in scrape_item.url.parts or "u" in scrape_item.url.parts:
                await self.user(scrape_item, reddit)
            elif "r" in scrape_item.url.parts and "comments" not in scrape_item.url.parts:
                await self.subreddit(scrape_item, reddit)
            elif "redd.it" in scrape_item.url.host:
                await self.media(scrape_item, reddit)
            else:
                await log(f"Scrape Failed: Unknown URL Path for {scrape_item.url}", 40)
                await self.manager.progress_manager.scrape_stats_progress.add_failure("Unknown")

        await self.scraping_progress.remove_task(task_id)

    @error_handling_wrapper
    async def user(self, scrape_item: ScrapeItem, reddit: asyncpraw.Reddit) -> None:
        """Scrapes user pages"""
        username = scrape_item.url.name
        title = await self.create_title(username, None, None)
        await scrape_item.add_to_parent_title(title)
        scrape_item.part_of_album = True

        user = await reddit.redditor(username)
        submissions = user.submissions.new(limit=None)
        await self.get_posts(scrape_item, submissions, reddit)

    @error_handling_wrapper
    async def subreddit(self, scrape_item: ScrapeItem, reddit: asyncpraw.Reddit) -> None:
        """Scrapes subreddit pages"""
        subreddit = scrape_item.url.name
        title = await self.create_title(subreddit, None, None)
        await scrape_item.add_to_parent_title(title)
        scrape_item.part_of_album = True

        subreddit = await reddit.subreddit(subreddit)
        submissions = subreddit.new(limit=None)
        await self.get_posts(scrape_item, submissions, reddit)

    @error_handling_wrapper
    async def get_posts(self, scrape_item: ScrapeItem, submissions, reddit: asyncpraw.Reddit) -> None:
        try:
            submissions = [submission async for submission in submissions]
        except asyncprawcore.exceptions.Forbidden:
            raise ScrapeFailure(403, "Forbidden")
        except asyncprawcore.exceptions.NotFound:
            raise ScrapeFailure(404, "Not Found")

        for submission in submissions:
            await self.post(scrape_item, submission, reddit)

    @error_handling_wrapper
    async def post(self, scrape_item: ScrapeItem, submission, reddit: asyncpraw.Reddit) -> None:
        """Scrapes posts"""
        title = submission.title
        date = int(str(submission.created_utc).split(".")[0])

        try:
            media_url = URL(submission.media['reddit_video']['fallback_url'])
        except (KeyError, TypeError):
            media_url = URL(submission.url)

        if "v.redd.it" in media_url.host:
            filename, ext = await get_filename_and_ext(media_url.name)

        if "redd.it" in media_url.host:
            new_scrape_item = await self.create_new_scrape_item(media_url, scrape_item, title, date)
            await self.media(new_scrape_item, reddit)
        elif "gallery" in media_url.parts:
            new_scrape_item = await self.create_new_scrape_item(media_url, scrape_item, title, date)
            await self.gallery(new_scrape_item, submission, reddit)
        else:
            if not "reddit.com" in media_url.host:
                new_scrape_item = await self.create_new_scrape_item(media_url, scrape_item, title, date)
                await self.handle_external_links(new_scrape_item)

    async def gallery(self, scrape_item: ScrapeItem, submission, reddit: asyncpraw.Reddit) -> None:
        """Scrapes galleries"""
        if not hasattr(submission, "media_metadata") or submission.media_metadata is None:
            return
        items = [item for item in submission.media_metadata.values() if item["status"] == "valid"]
        links = [URL(item["s"]["u"]).with_host("i.redd.it").with_query(None) for item in items]
        for link in links:
            new_scrape_item = await self.create_new_scrape_item(link, scrape_item, scrape_item.parent_title, scrape_item.possible_datetime)
            await self.media(new_scrape_item, reddit)

    @error_handling_wrapper
    async def media(self, scrape_item: ScrapeItem, reddit: asyncpraw.Reddit) -> None:
        """Handles media links"""
        try:
            filename, ext = await get_filename_and_ext(scrape_item.url.name)
        except NoExtensionFailure:
            head = await self.client.get_head(self.domain, scrape_item.url)
            head = await self.client.get_head(self.domain, head['location'])

            try:
                post = await reddit.submission(url=head['location'])
            except asyncprawcore.exceptions.Forbidden:
                raise ScrapeFailure(403, "Forbidden")
            except asyncprawcore.exceptions.NotFound:
                raise ScrapeFailure(404, "Not Found")

            await self.post(scrape_item, post, reddit)
            return

        await self.handle_file(scrape_item.url, scrape_item, filename, ext)

    """~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""

    async def create_new_scrape_item(self, link: URL, old_scrape_item: ScrapeItem, title: str, date: int) -> ScrapeItem:
        """Creates a new scrape item with the same parent as the old scrape item"""

        new_scrape_item = await self.create_scrape_item(old_scrape_item, link, "", True, date)
        if self.manager.config_manager.settings_data['Download_Options']['separate_posts']:
            await new_scrape_item.add_to_parent_title(title)
        return new_scrape_item



import asyncio
import copy
import shutil
from dataclasses import field
from pathlib import Path
from time import sleep
from typing import Dict, List, TYPE_CHECKING

import yaml

from cyberdrop_dl.managers.log_manager import LogManager
from cyberdrop_dl.utils.args.config_definitions import authentication_settings, settings, global_settings

if TYPE_CHECKING:
    from cyberdrop_dl.managers.manager import Manager


def _match_config_dicts(default: Dict, existing: Dict) -> Dict:
    """Matches the keys of two dicts and returns the default dict with the values of the existing dict"""
    for group in default:
        for key in default[group]:
            if group in existing and key in existing[group]:
                default[group][key] = existing[group][key]
    return default


def _save_yaml(file: Path, data: Dict) -> None:
    """Saves a dict to a yaml file"""
    file.parent.mkdir(parents=True, exist_ok=True)
    with open(file, 'w') as yaml_file:
        yaml.dump(data, yaml_file)


def _load_yaml(file: Path) -> Dict:
    """Loads a yaml file and returns it as a dict"""
    with open(file, 'r') as yaml_file:
        return yaml.load(yaml_file.read(), Loader=yaml.FullLoader)


class ConfigManager:
    def __init__(self, manager: 'Manager'):
        self.manager = manager
        self.loaded_config: str = field(init=False)

        self.authentication_settings: Path = field(init=False)
        self.settings: Path = field(init=False)
        self.global_settings: Path = field(init=False)

        self.authentication_data: Dict = field(init=False)
        self.settings_data: Dict = field(init=False)
        self.global_settings_data: Dict = field(init=False)

    def startup(self) -> None:
        """Pre-startup process for the config manager"""
        if not isinstance(self.loaded_config, str):
            self.loaded_config = self.manager.cache_manager.get("default_config")
            if not self.loaded_config:
                self.loaded_config = "Default"
            if self.manager.args_manager.load_config_from_args:
                self.loaded_config = self.manager.args_manager.load_config_name

        self.authentication_settings = self.manager.path_manager.config_dir / "authentication.yaml"
        self.global_settings = self.manager.path_manager.config_dir / "global_settings.yaml"
        self.settings = self.manager.path_manager.config_dir / self.loaded_config / "settings.yaml"

        self.settings.parent.mkdir(parents=True, exist_ok=True)
        self.load_configs()

    def load_configs(self) -> None:
        """Loads all the configs"""
        if self.authentication_settings.is_file():
            self._verify_authentication_config()
        else:
            self.authentication_data = copy.deepcopy(authentication_settings)
            _save_yaml(self.authentication_settings, self.authentication_data)

        if self.global_settings.is_file():
            self._verify_global_settings_config()
        else:
            self.global_settings_data = copy.deepcopy(global_settings)
            _save_yaml(self.global_settings, self.global_settings_data)

        if self.manager.args_manager.config_file:
            self.settings = Path(self.manager.args_manager.config_file)
            self.loaded_config = "CLI-Arg Specified"

        if self.settings.is_file():
            self._verify_settings_config()
        else:
            from cyberdrop_dl.managers.path_manager import APP_STORAGE, DOWNLOAD_STORAGE
            self.settings_data = copy.deepcopy(settings)
            self.settings_data['Files']['input_file'] = APP_STORAGE / "Configs" / self.loaded_config / "URLs.txt"
            self.settings_data['Files']['download_folder'] = DOWNLOAD_STORAGE / "Cyberdrop-DL Downloads"
            self.settings_data["Logs"]["log_folder"] = APP_STORAGE / "Configs" / self.loaded_config / "Logs"
            self.settings_data['Sorting']['sort_folder'] = DOWNLOAD_STORAGE / "Cyberdrop-DL Sorted Downloads"
            self.write_updated_settings_config()

    def _verify_authentication_config(self) -> None:
        """Verifies the authentication config file and creates it if it doesn't exist"""
        default_auth_data = copy.deepcopy(authentication_settings)
        existing_auth_data = _load_yaml(self.authentication_settings)
        self.authentication_data = _match_config_dicts(default_auth_data, existing_auth_data)
        _save_yaml(self.authentication_settings, self.authentication_data)

    def _verify_settings_config(self) -> None:
        """Verifies the settings config file and creates it if it doesn't exist"""
        default_settings_data = copy.deepcopy(settings)
        existing_settings_data = _load_yaml(self.settings)
        self.settings_data = _match_config_dicts(default_settings_data, existing_settings_data)
        self.settings_data['Files']['input_file'] = Path(self.settings_data['Files']['input_file'])
        self.settings_data['Files']['download_folder'] = Path(self.settings_data['Files']['download_folder'])
        self.settings_data["Logs"]["log_folder"] = Path(self.settings_data["Logs"]["log_folder"])
        self.settings_data['Sorting']['sort_folder'] = Path(self.settings_data['Sorting']['sort_folder'])

        # change to ints
        self.settings_data['File_Size_Limits']['maximum_image_size'] = int(self.settings_data['File_Size_Limits']['maximum_image_size'])
        self.settings_data['File_Size_Limits']['maximum_video_size'] = int(self.settings_data['File_Size_Limits']['maximum_video_size'])
        self.settings_data['File_Size_Limits']['maximum_other_size'] = int(self.settings_data['File_Size_Limits']['maximum_other_size'])
        self.settings_data['File_Size_Limits']['minimum_image_size'] = int(self.settings_data['File_Size_Limits']['minimum_image_size'])
        self.settings_data['File_Size_Limits']['minimum_video_size'] = int(self.settings_data['File_Size_Limits']['minimum_video_size'])
        self.settings_data['File_Size_Limits']['minimum_other_size'] = int(self.settings_data['File_Size_Limits']['minimum_other_size'])

        self.global_settings_data['General']['max_file_name_length'] = int(self.global_settings_data['General']['max_file_name_length'])
        self.global_settings_data['General']['max_folder_name_length'] = int(self.global_settings_data['General']['max_folder_name_length'])
        self.global_settings_data['Rate_Limiting_Options']['connection_timeout'] = int(self.global_settings_data['Rate_Limiting_Options']['connection_timeout'])
        self.global_settings_data['Rate_Limiting_Options']['download_attempts'] = int(self.global_settings_data['Rate_Limiting_Options']['download_attempts'])
        self.global_settings_data['Rate_Limiting_Options']['download_delay'] = int(self.global_settings_data['Rate_Limiting_Options']['download_delay'])
        self.global_settings_data['Rate_Limiting_Options']['max_simultaneous_downloads'] = int(self.global_settings_data['Rate_Limiting_Options']['max_simultaneous_downloads'])
        self.global_settings_data['Rate_Limiting_Options']['max_simultaneous_downloads_per_domain'] = int(self.global_settings_data['Rate_Limiting_Options']['max_simultaneous_downloads_per_domain'])
        self.global_settings_data['Rate_Limiting_Options']['rate_limit'] = int(self.global_settings_data['Rate_Limiting_Options']['rate_limit'])
        self.global_settings_data['Rate_Limiting_Options']['read_timeout'] = int(self.global_settings_data['Rate_Limiting_Options']['read_timeout'])

        save_data = copy.deepcopy(self.settings_data)
        save_data['Files']['input_file'] = str(save_data['Files']['input_file'])
        save_data['Files']['download_folder'] = str(save_data['Files']['download_folder'])
        save_data["Logs"]["log_folder"] = str(save_data["Logs"]["log_folder"])
        save_data['Sorting']['sort_folder'] = str(save_data['Sorting']['sort_folder'])
        _save_yaml(self.settings, save_data)

    def _verify_global_settings_config(self) -> None:
        """Verifies the global settings config file and creates it if it doesn't exist"""
        default_global_settings_data = copy.deepcopy(global_settings)
        existing_global_settings_data = _load_yaml(self.global_settings)
        self.global_settings_data = _match_config_dicts(default_global_settings_data, existing_global_settings_data)
        _save_yaml(self.global_settings, self.global_settings_data)

    """~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""

    def create_new_config(self, new_settings: Path, settings_data: Dict) -> None:
        """Creates a new settings config file"""
        settings_data['Files']['input_file'] = str(settings_data['Files']['input_file'])
        settings_data['Files']['download_folder'] = str(settings_data['Files']['download_folder'])
        settings_data["Logs"]["log_folder"] = str(settings_data["Logs"]["log_folder"])
        settings_data['Sorting']['sort_folder'] = str(settings_data['Sorting']['sort_folder'])
        _save_yaml(new_settings, settings_data)

    def write_updated_authentication_config(self) -> None:
        """Write updated authentication data"""
        _save_yaml(self.authentication_settings, self.authentication_data)

    def write_updated_settings_config(self) -> None:
        """Write updated settings data"""
        settings_data = copy.deepcopy(self.settings_data)
        settings_data['Files']['input_file'] = str(settings_data['Files']['input_file'])
        settings_data['Files']['download_folder'] = str(settings_data['Files']['download_folder'])
        settings_data["Logs"]["log_folder"] = str(settings_data["Logs"]["log_folder"])
        settings_data['Sorting']['sort_folder'] = str(settings_data['Sorting']['sort_folder'])
        _save_yaml(self.settings, settings_data)

    def write_updated_global_settings_config(self) -> None:
        """Write updated global settings data"""
        _save_yaml(self.global_settings, self.global_settings_data)

    """~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""

    def get_configs(self) -> List:
        """Returns a list of all the configs"""
        return [config.name for config in self.manager.path_manager.config_dir.iterdir() if config.is_dir()]

    def change_default_config(self, config_name: str) -> None:
        """Changes the default config"""
        self.manager.cache_manager.save("default_config", config_name)

    def delete_config(self, config_name: str) -> None:
        """Deletes a config"""
        configs = self.get_configs()
        configs.remove(config_name)

        if self.manager.cache_manager.get("default_config") == config_name:
            self.manager.cache_manager.save("default_config", configs[0])

        config = self.manager.path_manager.config_dir / config_name
        shutil.rmtree(config)

    def change_config(self, config_name: str) -> None:
        """Changes the config"""
        self.loaded_config = config_name
        self.startup()

        self.manager.path_manager.startup()
        sleep(1)
        self.manager.log_manager = LogManager(self.manager)
        sleep(1)
